# type: ignore
"""Test page.py."""

from os.path import join

import pytest
from . import page
from .utils import check_requirements

SQL = """
    PRAGMA foreign_keys=ON;
    INSERT INTO Files (filename) VALUES ('test.md');
    INSERT INTO Notes (id, title, filename) VALUES
        (0, '0', 'test.md'),
        (1, '1', 'test.md'),
        (2, '2', 'test.md');
    INSERT INTO Links (tag, src, dest) VALUES
        ('#test', 0, 0),
        ('#test', 1, 1),
        ('##test', 2, 2);
    INSERT INTO Bibliography (key, text) VALUES
        ('ref-test', 'Reference text.');
    INSERT INTO Citations (note, reference) VALUES
        (0, 'ref-test');
"""

def test_data_path():
    """Check if path is constructed correctly."""
    path = page.data_path("frontend.js")
    assert join("data", "frontend.js") in str(path)

def test_data_shell_path():
    """Check if path is constructed and escaped correctly."""
    path = page.data_shell_path("frontend.js")
    assert join("data", "frontend.js") in str(path)

    path = page.data_shell_path("front end.js")
    assert path[0] == "'"
    assert join("data", "front end.js") + "'" in path

def test_create_bibliography(mock_db):
    """Check create_bibliography output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_bibliography(conn)
    assert html == \
"""<section id='references' title='References' class='level1'>
  <h1>
    References
  </h1>
  <dl>
    <dt>
      <a href='#ref-test'>
        [@test]
      </a>
    </dt>
    <dd>
      Reference text.
    </dd>
  </dl>
</section>"""

def test_create_tags(mock_db):
    """Check create_tags output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_tags(conn)
    assert html == """<section id='tags' title='Tags' class='level1'>
  <h1>
    Tags
  </h1>
  <ul>
    <li>
      <a href='###test'>
        ##test
      </a>
       (1)
    </li>
    <li>
      <a href='##test'>
        #test
      </a>
       (2)
    </li>
  </ul>
</section>"""

def test_create_tag_page(mock_db):
    """Check create_tag_page output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_tag_page(conn, "#test")
    assert html == """<section id='#test' title='#test' class='level1'>
  <h1>
    <a href='#tags' title='List of tags'>
      #test
    </a>
  </h1>
  <ol class='slipbox-list'>
    <li value='0'></li>
    <li value='1'></li>
  </ol>
</section>"""

def test_create_reference_page(mock_db):
    """Check create_reference_page output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.create_reference_page(conn, "ref-test")
    assert html == """<section id='ref-test' title='ref-test' class='level1'>
  <h1>
    <a href='#references'>
      @test
    </a>
  </h1>
  <p>
    Reference text.
  </p>
  <ol class='slipbox-list'>
    <li value='0'></li>
  </ol>
</section>"""

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_generate_complete_html(mock_db, tmp_path):
    """Sanity check."""
    options = ""
    page.generate_complete_html(mock_db, options, tmp_path)



"""slipbox CLI commands."""

from pathlib import Path
import sys
from typing import Optional, Sequence

from . import check
from .initializer import DotSlipbox, default_config
from .slipbox import Slipbox

def require_dot_slipbox(path: Path = Path()) -> DotSlipbox:
    """Return .slipbox object in current directory.

    Exit if it does not exist.
    """
    dot = DotSlipbox.locate(path.resolve())
    if dot is None:
        sys.exit("could not find '.slipbox' in any parent directory.")
    return dot

def show_info(dot: DotSlipbox, note_id: int) -> None:
    """Print metadata associated with note ID."""
    sql = "SELECT title, filename FROM Notes WHERE id = ?"
    with Slipbox(dot) as slipbox:
        cur = slipbox.conn.cursor()
        cur.execute(sql, (note_id,))
        note = cur.fetchone()
        if note is not None:
            print(note_id)
            print(note[0])
            print(note[1])

def show_info_wrapper(note_id: str, /) -> None:
    """Show information about note."""
    dot = require_dot_slipbox()
    return show_info(dot, int(note_id))

def main() -> None:
    """Compile notes into static page."""
    dot = require_dot_slipbox()
    with Slipbox(dot) as slipbox:
        slipbox.run()

def check_notes() -> None:
    """Check notes in slipbox for invalid links and isolated notes."""
    dot = require_dot_slipbox()
    with Slipbox(dot) as slipbox:
        if not check.check_notes(slipbox):
            sys.exit(65)

def generate_flashcards(output: str, /) -> None:
    """Generate anki flash cards from notes."""
    try:
        import genanki #type: ignore
    except ImportError:
        sys.exit("could not import genanki")
    model = genanki.Model(
        1635490798, # model ID
        "Note",
        fields=[
            {"name": "Question"},
            {"name": "Answer"},
        ],
        templates=[
            {
                "name": "Card",
                "qfmt": "<h1>{{Question}}</h1>",
                "afmt": "{{Answer}}",
            },
        ],
    )
    deck = genanki.Deck(1843743983, # deck ID
                        "Slipbox")
    dot = require_dot_slipbox()
    with Slipbox(dot) as slipbox:
        sql = "SELECT title, html FROM Notes"
        for title, html in slipbox.conn.execute(sql):
            note = genanki.Note(model=model,
                                fields=[
                                    title,
                                    html.replace(' style="display:none"', "")
                                ])
            deck.add_note(note)
    genanki.Package(deck).write_to_file(Path(output))

def initialize(directory: Optional[str] = None,
               /,
               content_options: Optional[str] = None,
               document_options: Optional[str] = None) -> None:
    """Initialize notes directory."""
    parent = Path(directory) if directory else Path()
    parent.mkdir(parents=True, exist_ok=True)

    defaults = default_config()
    if not content_options:
        content_options = defaults.get("slipbox", "content_options")
    if not document_options:
        document_options = defaults.get("slipbox", "document_options")

    DotSlipbox(parent, dict(content_options=content_options, document_options=document_options))
    print(f"Initialized .slipbox in {parent.resolve()!s}.")

def has_gaps(sequence: Sequence[int]) -> bool:
    """Check if sequence has gaps.

    Assume sequence is an increasing sequence of non-negative integers with no
    duplicate entries.
    """
    return bool(sequence) and (sequence[-1] - sequence[0] >= len(sequence))

def find_available_id(sequence: Sequence[int]) -> int:
    """Return smallest non-negative integer not in the sequence."""
    if not sequence or sequence[0] > 0:
        return 0
    if not has_gaps(sequence):
        return sequence[-1] + 1
    while len(sequence) > 2:
        mid = len(sequence) // 2
        lower = sequence[:mid]
        boundary = sequence[mid-1:mid+1]
        upper = sequence[mid:]
        sequence = lower if has_gaps(lower) \
            else boundary if has_gaps(boundary) \
            else upper
    return sequence[0] + 1

def new_note(note_format: Optional[str] = None, /) -> None:
    """Get smallest available note ID for new note."""
    dot = require_dot_slipbox()
    with Slipbox(dot) as slipbox:
        rows = slipbox.conn.execute("SELECT id FROM Notes ORDER BY (id)")
        ids = [row[0] for row in rows]
    suggest = find_available_id(ids)
    if note_format == "markdown":
        print(f"# {suggest} New note")
    else:
        print(suggest)



#type: ignore
"""Test secparse.py."""

from .secparse import note_id, parse_sections

def test_note_id():
    """Test note_id on different inputs.."""
    assert note_id("div", [("id", "1")]) is None
    assert note_id("section", [("id", None)]) is None
    assert note_id("section", [("id", "-1")]) is None
    assert note_id("section", [("id", "42")]) == 42

def test_parse_sections():
    """Test parse_sections."""
    section1 = '<section id="1">&gt;<!--comment--><hr /></section>'
    section2 = '<section id="2"><section>&#x3E;</section></section>'
    test_html = section1 + '\n' + section2

    result = []
    def callback(this):
        result.append((this.id_, this.section.strip()))
    parse_sections(test_html, callback)

    assert result == [(1, section1), (2, section2)]



"""Static site generator for Zettelkasten notes."""

import sys
from .cli import cli
from .utils import check_requirements

if __name__ == "__main__":
    if not check_requirements():
        sys.exit("[ERROR] pandoc not found.")
    cli.run()



"""Check slipbox notes."""

from typing import Iterator, Tuple
from .slipbox import Slipbox
from .utils import print_sequence

_Note = Tuple[int, str, str]

def invalid_links(slipbox: Slipbox) -> Iterator[Tuple[_Note, int]]:
    """Generate notes that link to invalid ID."""
    sql = """
        SELECT DISTINCT id, title, filename, dest
        FROM Links JOIN Notes ON src = id
        WHERE dest NOT IN (
            SELECT id FROM Notes
        )
        ORDER BY id
    """
    for nid, title, filename, dest in slipbox.conn.execute(sql):
        yield (nid, title, filename), dest

def isolated_notes(slipbox: Slipbox) -> Iterator[_Note]:
    """Generate isolated notes (untagged)."""
    yield from slipbox.conn.execute("""
        SELECT DISTINCT id, title, filename FROM Notes
        WHERE id NOT IN (
            SELECT src FROM Links UNION SELECT dest FROM Links
        )
    """)

def unsourced_notes(slipbox: Slipbox) -> Iterator[_Note]:
    """Generate notes that need citations (only if there's a bibliography)."""
    if "--bibliography" in slipbox.config["slipbox"]["content_options"]:
        yield from slipbox.conn.execute("""
            SELECT DISTINCT id, title, filename FROM Notes
            WHERE id NOT IN (
                SELECT note FROM Citations
            )
        """)

def check_notes(slipbox: Slipbox) -> bool:
    """Check notes in slipbox.

    Returns false is errors are found.
    """
    format_note = lambda note: f"  {note[0]}. {note[1]} in {note[2]!r}."
    format_link = lambda x: f"  {x[0][0]}. {x[0][1]} in {x[0][2]!r} -> {x[1]}."
    _invalid_links = invalid_links(slipbox)
    _isolated_notes = isolated_notes(slipbox)
    _unsourced_notes = unsourced_notes(slipbox)

    errors = [
        print_sequence("The following notes link to non-existent notes.",
                       map(format_link, _invalid_links)),
        print_sequence("The following notes are not connected to other notes.",
                       map(format_note, _isolated_notes)),
        print_sequence("The following notes have missing citations.",
                       map(format_note, _unsourced_notes)),
    ]
    return not any(errors)



# type: ignore
"""Test app.py."""

import pytest

from . import app
from .initializer import DotSlipbox
from .slipbox import Slipbox
from .utils import check_requirements

def test_show_info_timestamp_empty(tmp_path):
    """Calling show_info shouldn't affect the db timestamp."""
    dot = DotSlipbox(tmp_path)
    with Slipbox(dot) as slipbox:
        before = slipbox.timestamp
    app.show_info(dot, 0)
    with Slipbox(dot) as slipbox:
        after = slipbox.timestamp
    assert before == after

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_show_info_timestamp(tmp_path, mnote, sbox):
    """Calling show_info shouldn't affect the db timestamp."""
    sbox.process([mnote])
    dot = DotSlipbox(tmp_path)
    with Slipbox(dot) as slipbox:
        before = slipbox.timestamp
    app.show_info(dot, 0)
    with Slipbox(dot) as slipbox:
        after = slipbox.timestamp
    assert before == after

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_show_info_empty_stdout(tmp_path, capsys, mnote, sbox):
    """show_info shouldn't output anything if the note ID doesn't exist."""
    sbox.process([mnote])
    dot = DotSlipbox(tmp_path)
    app.show_info(dot, 1)
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_show_info_stdout(tmp_path, capsys, mnote, sbox):
    """show_info should output info of note with give note ID."""
    sbox.process([mnote])
    dot = DotSlipbox(tmp_path)
    app.show_info(dot, 0)
    stdout, stderr = capsys.readouterr()
    assert stdout
    assert not stderr

def test_require_dot_slipbox(tmp_path, sbox):
    """require_dot_slipbox should return DotSlipbox object in current directory."""
    dot = app.require_dot_slipbox(tmp_path)
    assert dot.path == sbox.dot.path

def test_require_dot_slipbox_uninitialized(tmp_path):
    """require_dot_slipbox should abort the program."""
    with pytest.raises(SystemExit):
        app.require_dot_slipbox(tmp_path)

def test_has_gaps():
    """has_gaps should check if input has gaps."""
    assert not app.has_gaps([])
    assert not app.has_gaps([1])
    assert not app.has_gaps([2, 3])
    assert not app.has_gaps([4, 5, 6])

    assert app.has_gaps([7, 9])
    assert app.has_gaps([0, 2])

def test_find_available_id():
    """find_available_id should return smallest non-negative integer not in the input."""
    find = app.find_available_id
    assert find([]) == 0
    assert find([0]) == 1
    assert find([1]) == 0
    assert find(list(range(1, 10))) == 0
    assert find(list(range(10))) == 10
    assert find([0, 2]) == 1
    assert find([0, 1, 3]) == 2



# type: ignore
"""Test utils.py."""

import shlex
import shutil

import pytest

from . import utils

def test_sqlite_string():
    """Single quotes must be escaped properly."""
    assert utils.sqlite_string("''") == "''''''"
    assert utils.sqlite_string("'foo'bar'") == "'''foo''bar'''"

def test_temporary_directory():
    """temporary_directory gets deleted when the context manager exits."""
    with utils.temporary_directory() as temp:
        path = temp
        assert temp.exists()
        assert temp.is_dir()
    assert not path.exists()

@pytest.mark.skipif(not shutil.which("grep"), reason="requires grep")
def test_run_command(tmp_path, capsys):
    """run_command must return returncode and capture stdout and stderr."""
    first = tmp_path/"first.txt"
    second = tmp_path/"second.txt"
    first.write_text("first")
    second.write_text("second")

    retcode = utils.run_command("grep first {} {}".format(
        shlex.quote(str(first)), shlex.quote(str(second))
    ))

    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert not retcode
    assert str(first) in stdout
    assert str(second) not in stdout

    retcode = utils.run_command("grep second {} {}".format(
        shlex.quote(str(first)), shlex.quote(str(second))
    ))
    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert str(first) not in stdout
    assert str(second) in stdout

@pytest.mark.skipif(not shutil.which("env"), reason="requires env")
def test_run_command_with_kwargs(capsys):
    """Keyword arguments to run_command must be used as environment variables.
    """
    retcode = utils.run_command("env", dict(ZZZZZZZZZZ="ZZZZZZZZZZ"))
    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert not retcode
    assert "ZZZZZZZZZZ=ZZZZZZZZZZ" in stdout

def test_print_sequence_empty(capsys):
    """print_sequence must not print anything if sequence is empty.

    The result must be False (empty).
    """
    assert not utils.print_sequence("error", [])
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

def test_print_sequence_not_empty(capsys):
    """print_sequence must print header if sequence is not empty.

    The result must be True (non-empty)."""
    assert utils.print_sequence("hello", ["world", "!"])
    stdout, stderr = capsys.readouterr()
    assert stdout
    assert "hello" in stdout
    assert "world" in stdout
    assert "!" in stdout
    assert not stderr



"""Generate complete HTML containing all slipbox notes."""

from pathlib import Path
import shlex
from sqlite3 import Connection
import subprocess
from typing import Iterable

from .templates import Elem, render
from .utils import pandoc, temporary_directory

DUMMY_MARKDOWN = r"""::: {style="display:none"}
$\,$
```c
```
:::"""

def data_path(filename: str) -> Path:
    """Return path in data directory."""
    return Path(__file__).parent/"data"/filename

def data_shell_path(filename: str) -> str:
    """Return shell-escaped filename in data directory."""
    return shlex.quote(str(data_path(filename)))

def generate_active_htmls(conn: Connection) -> Iterable[str]:
    """Get HTML stored in the database for active sections."""
    sql = "SELECT html FROM Notes WHERE html IS NOT NULL ORDER BY id ASC"
    return (html.strip() for html, in conn.execute(sql))

def generate_data(conn: Connection) -> Iterable[str]:
    """Generate slipbox data in javascript."""
    yield "for (const sec of document.querySelectorAll('section.slipbox-note')) {"
    yield "  window.slipbox.addNote(sec.id)"
    yield "}"
    sql = "SELECT src, dest, tag FROM ValidLinks ORDER BY src, dest, tag"
    for src, dest, tag in conn.execute(sql):
        tag = "null" if not tag else repr(tag)
        yield f"window.slipbox.addLink({str(src)!r}, {str(dest)!r}, {tag})"

def generate_javascript(conn: Connection) -> Iterable[str]:
    """Generate slipbox javascript code."""
    yield '<script type="module">'
    yield data_path("frontend.js").read_text().strip()
    yield '</script>'
    yield '<script>'
    yield "window.addEventListener('DOMContentLoaded', () => {"
    yield from generate_data(conn)
    yield "window.initSlipbox()"
    yield "})"
    yield "</script>"

def create_bibliography(conn: Connection) -> str:
    """Create bibliography HTML section from database entries."""
    sql = "SELECT key, text FROM Bibliography ORDER BY key"
    items = []
    for key, text in conn.execute(sql):
        items.append(Elem("dt", Elem("a", f"[@{key[4:]}]", href='#' + key)))
        items.append(Elem("dd", text))
    section = Elem("section",
                   Elem("h1", "References"),
                   Elem("dl", *items),
                   id="references",
                   title="References",
                   **{"class": "level1"})
    return render(section)

def create_tags(conn: Connection) -> str:
    """Create HTML section that lists all tags.

    Also list untagged notes.
    """
    rows = conn.execute("SELECT tag, COUNT(*) FROM Tags GROUP BY tag ORDER BY tag")
    items = (Elem("li", Elem("a", tag, href=f"#{tag}"), f" ({count})") for tag, count in rows)
    section = Elem("section",
                   Elem("h1", "Tags"),
                   Elem("ul", *items),
                   id="tags",
                   title="Tags",
                   **{"class": "level1"})
    untagged = list(conn.execute("SELECT * FROM Untagged"))
    if untagged:
        section.children.append(Elem("h2", "Untagged notes"))
        items = (Elem("li", value=str(nid)) for nid, in untagged)
        section.children.append(Elem("ol", *items, **{"class": "slipbox-list"}))
    return render(section)

def create_tag_page(conn: Connection, tag: str) -> str:
    """Create HTML section that lists all notes with the tag."""
    sql = """
        SELECT id FROM Tags NATURAL JOIN Notes WHERE tag = ? ORDER BY id
    """
    items = []
    for nid, in conn.execute(sql, (tag,)):
        item = Elem("li", value=str(nid))
        items.append(item)
    section = Elem("section",
                   Elem("h1", Elem("a", tag, href="#tags", title="List of tags")),
                   Elem("ol", *items, **{"class": "slipbox-list"}),
                   id=tag,
                   title=tag,
                   **{"class": "level1"})
    return render(section)

def create_tag_pages(conn: Connection) -> str:
    """Create all tag pages."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    return '\n'.join(create_tag_page(conn, tag) for tag in tags)

def create_reference_page(conn: Connection, reference: str) -> str:
    """Create HTML section that lists all notes that cite the reference."""
    sql = """
        SELECT note, text FROM Citations
            JOIN Notes ON Citations.note = Notes.id
                JOIN Bibliography ON Bibliography.key = Citations.reference
                    WHERE reference = ?
                        ORDER BY note
    """
    items = []
    text = ""
    for note, _text in conn.execute(sql, (reference,)):
        text = _text
        item = Elem("li", value=str(note))
        items.append(item)
    section = Elem("section",
                   Elem("h1", Elem("a", '@' + reference[4:], href="#references")),
                   Elem("p", text),
                   Elem("ol", *items, **{"class": "slipbox-list"}),
                   id=reference,
                   title=reference,
                   **{"class": "level1"})
    return render(section)

def create_reference_pages(conn: Connection) -> str:
    """Create all reference pages."""
    rows = conn.execute("SELECT key FROM Bibliography ORDER BY key")
    references = (row[0] for row in rows)
    return '\n'.join(create_reference_page(conn, ref) for ref in references)

def generate_complete_html(conn: Connection, options: str, basedir: Path) -> None:
    """Create final HTML file with javascript."""
    with temporary_directory() as tempdir:
        script = tempdir/"script.js"
        html = tempdir/"cached.html"
        extra = tempdir/"extra.html"
        dummy = tempdir/"Slipbox.md"
        dummy.write_text(DUMMY_MARKDOWN, encoding="utf-8")
        script.write_text('\n'.join(generate_javascript(conn)), encoding="utf-8")
        html.write_text('\n'.join(generate_active_htmls(conn)), encoding="utf-8")
        with open(extra, "w", encoding="utf-8") as file:
            print(create_tag_pages(conn), file=file)
            print(create_tags(conn), file=file)
            print(create_reference_pages(conn), file=file)
            print(create_bibliography(conn), file=file)
        cmd = """{pandoc} {dummy} -H{script} {title} -A{nav} -A{html} -A{extra} -A{search}
                -A{bottom} --section-divs {opts} -H {style}
            """.format(
            pandoc=pandoc(), dummy=shlex.quote(str(dummy)), script=shlex.quote(str(script)),
            html=shlex.quote(str(html)), opts=options, extra=shlex.quote(str(extra)),
            title="--metadata title=Slipbox",
            style=data_shell_path("style.html"),
            nav=data_shell_path("nav.html"),
            bottom=data_shell_path("bottom.html"),
            search=data_shell_path("search.html"))
        subprocess.run(shlex.split(cmd), check=False, cwd=basedir)
